// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html.

#include "precomp.hpp"

// see modules/video/src/tracking/tracker.cpp

#include "legacy/tracker.legacy.hpp"
