// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html.
#ifndef __OPENCV_DNN_SUPERRES_PRECOMP_HPP__
#define __OPENCV_DNN_SUPERRES_PRECOMP_HPP__

#include <iostream>
#include <vector>
#include <memory>
#include <string>
#include <map>
#include <numeric>
#include <algorithm>

#include "opencv2/core.hpp"
#include <opencv2/core/utils/logger.hpp>

#include "opencv2/dnn.hpp"
#include "opencv2/imgproc.hpp"

#include "opencv2/dnn_superres.hpp"

#endif // __OPENCV_DNN_SUPERRES_PRECOMP_HPP__
