// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html
// Copyright (C) 2020 by Archit Rungta

// Needed to prevent documentation warning

namespace cv
{
    namespace julia
    {
        void initJulia(int, char **) {}
    } // namespace julia
} // namespace cv
