// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html.

#ifndef OPENCV_INTENSITY_TRANSFORM_PRECOMP_H
#define OPENCV_INTENSITY_TRANSFORM_PRECOMP_H

#include "opencv2/core.hpp"
#include "opencv2/intensity_transform.hpp"
#include "opencv2/core/private.hpp"
#include <array>

#endif
