// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html.
#ifndef OPENCV_QUALITY_PRECOMP_HPP
#define OPENCV_QUALITY_PRECOMP_HPP
#include <opencv2/core.hpp>
#include "opencv2/quality/qualitybase.hpp"
#endif