// This file is part of OpenCV project.
// It is subject to the license terms in the LICENSE file found in the top-level directory
// of this distribution and at http://opencv.org/license.html

// This code is also subject to the license terms in the LICENSE_KinectFusion.md file found in this module's directory

// This code is also subject to the license terms in the LICENSE_WillowGarage.md file found in this module's directory

#ifndef __OPENCV_TEST_PRECOMP_HPP__
#define __OPENCV_TEST_PRECOMP_HPP__

#include <opencv2/ts.hpp>
#include <opencv2/rgbd.hpp>
#include <opencv2/calib3d.hpp>

#ifdef HAVE_OPENCL
#include <opencv2/core/ocl.hpp>
#endif

namespace opencv_test {
using namespace cv::rgbd;
}

#endif
